#ifndef THRESH_H
#define THRESH_H
#include <vector>
#include <map>
#include <string>
#include <iostream>
#include <fstream>
#include <eigen3/Eigen/Dense>
#include <random>
#include <exception>
#include <boost/math/distributions/normal.hpp> // for normal_distribution
    using boost::math::normal; // typedef provides default type is normal
#include <boost/math/distributions/beta.hpp>
    using boost::math::beta_distribution;

using namespace Eigen;
#include "qfc.h"

namespace thresh
{
    enum SIM_TYPE
    {
       BURDEN_ONLY,
       SKAT_ONLY,
       BURDEN_AND_SKAT
    };

    enum subject_data_type
    {
        FULL_DATA,           
        MISSING_DATA               
    };

      const double DEF_ALPHA_PARAM = 1.0;               
      const double DEF_BETA_PARAM = 25.0;

      int read_configfile(
        const std::string & config_filename,
              std::string & geno_filename,
              std::string & out_filename,
                   size_t & start_rep,
                   size_t & num_reps,
                   double & MAF_cutoff,
                   size_t & min_MAC,
                 bool & use_pheno,
                 std::string & pheno_filename,
                 SIM_TYPE & sim_type,
                   size_t & window_size,
                   size_t & window_step,
                   bool & make_map_from_window,
                   std::string & map_filename,
                   size_t & out_freq,
                   double & min_call_rate  );


        //****************************
        //Methods for simulations with covariates and phenotype
        void calc_U_vec_linear(
            double sigma_sqr,
            const VectorXd & gamma,
            const MatrixXd & G_mat,
            const MatrixXd & X_mat,
            const VectorXd & Y_vec,
                  VectorXd & U_vec );


        int calc_V_mat_linear(
            double sigma_sqr,
            const MatrixXd & G_mat,
            const MatrixXd & X_mat,
                  MatrixXd & V_mat );

        int calc_Burden_Q(
            const VectorXd & c_vec,
            const VectorXd & U_vec,
            const MatrixXd & V_mat,
                  double & Q_val  );


        int linear_sim_Y_vec_from_null(
            size_t seed_it,
            const MatrixXd & X_mat,
            const VectorXd & gamma, //from linear regression
                  double sigma_sqr, //from linear regression
                  VectorXd & Y_vec );

        int resample_Burden_threshold_linear(
			const MatrixXd & X_mat,
			const VectorXd & Y_vec,
			const std::vector< MatrixXd > & geno_vec,
			const std::string & outfilename,
				size_t start_rep = 0,
				size_t num_reps = 10000,
				size_t out_freq = 1000,
                size_t min_mac = 0);

       int perform_linear_reg(const MatrixXd & X_mat,
                   const VectorXd & Y_vec,
                   VectorXd & beta_vec,
                   VectorXd & se_vec,
                   double & sigma_sqr,
                   double & sigma_sqr_alt );

    //******************************************
   //file methods

     int read_map_file( const std::string & filename,
                   std::map< std::string, std::vector< std::string> > & snp_to_gene_map );	

     int read_pheno_data( std::string infilename,
                MatrixXd & X_mat,
                MatrixXd & Y_col_vec,
                std::vector< subject_data_type > & subject_data_vec
                );

     int read_geno_file( const std::string & filename,           
                    std::map<std::string, std::vector<std::string> > & snp_to_gene_map,
                    std::vector< MatrixXd > &  geno_mat_vec,
                    std::vector< std::string > & gene_names_vec,
                    std::vector< subject_data_type > & subject_data_vec_missing_pheno,
                    double min_call_rate 
                  ); 

     int make_window_map_from_geno( 
                    const std::string & geno_filename,
                    const std::string & map_filename,
                    size_t window_size,
                    size_t start_step_size );

    //************************************
    //no covariates and no phenotype methods.
    int linear_sim_Y_vec_nocovar(
                    size_t seed_it,
                    size_t NUM_CASES,
                    VectorXd & Y_vec );

    int resample_Burden_threshold_linear_nocovar(
            const std::vector< MatrixXd > & geno_vec,
            const std::string & outfilename,
            size_t start_rep = 0,
            size_t num_reps = 10000,
            size_t out_freq = 1000,
            size_t min_mac = 0
            );

    int calc_Burden_Q_linear_nocovar( const VectorXd & Y_vec,
                                    const VectorXd & sum_G_vxd,
                                    double & Q_val );


    //SKAT methods
    int resample_SKAT_threshold_linear_nocovar(
            const std::vector< MatrixXd > & geno_vec,
            const std::string & outfilename,
                  size_t start_rep = 0,   
                  size_t num_reps = 10000,
                  size_t out_freq = 1000,
                  size_t min_mac = 0 );	

     void calc_U_vec_linear_nocovar(  
        const MatrixXd & G_mat,
        const VectorXd & Y_vec,
              VectorXd & U_vec );

        int calc_V_mat_linear_nocovar(   
           const MatrixXd & G_mat,
                 MatrixXd & V_mat );


    void calc_MAF_vec( const MatrixXd & G_mat, VectorXd & MAF_vec );


    void calc_SKAT_W_mat(            
         const VectorXd & MAF_vec,
               MatrixXd & W_mat,
               double alpha = DEF_ALPHA_PARAM,
               double beta = DEF_BETA_PARAM );


    int calc_eigen_decomp(           
          const MatrixXd & A_mat,
                VectorXd & eigenvals_vec,
                MatrixXd & eigenvecs_mat );

    int calc_SKAT_pval(
                const VectorXd & U_vec,
                const MatrixXd & V_mat,
                const MatrixXd & G_mat,
                      double & pval,
                      double alpha = DEF_ALPHA_PARAM,
                      double beta = DEF_BETA_PARAM);

        int resample_SKAT_threshold_linear(
            const MatrixXd & X_mat,      
            const VectorXd & Y_vec,                  
            const std::vector< MatrixXd > & geno_vec,
            const std::string & outfilename,
                  size_t start_rep = 0,   
                  size_t num_reps = 10000,
                  size_t out_freq = 1000,
                  size_t min_mac = 0 );

        int calc_Burden_c_vec(           
            const MatrixXd & G_mat,
                  double prob_cut_off,
                  VectorXd & c_vec );


        int calc_above_min_MAF_c_vec(
            const MatrixXd & G_mat,
                  double min_MAF,
                  VectorXd & c_vec );


    //************************************* 

    //return values
    const int NO_ERROR  = 0;
    const int DIM_ERROR = 2;
    const int ARG_ERROR = 4;
    const int SIM_ERROR = 8;
    const int FILE_ERROR = 16;
    const int INVERT_ERROR = 32;
    const int SOLVER_ERROR = 64;
}
#endif

